/******************************************************************************

  (c) 2009-2011 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "MSG_Buffer_Handle.h"

/// Allocate buffer
void MSG_Buffer_Handle::alloc()
{
  /// Assign working buffer and push it to the container of buffers
  working_buf = malloc(max_size);
  buffers.push_back(working_buf);

  /// Reset the initial buffer parameters and increment number of buffers
  cur_buf = (void*)((char*)working_buf + 2*INT_SIZE);
  buf_size = 2*INT_SIZE;
  num_blocks = 0;
  num_buffers++;
  num_active_buffers++;
}

/// Find available buffer which can be used for putting messages in
void MSG_Buffer_Handle::get_buffer_to_send(int rep_num_buffer)
{
  /// If no buffer is available, allocate another one
  if (rep_num_buffer == -1)
  {
    alloc();
    return;
  }
  else
  {
    int last_ind = num_active_buffers - 1;

    /// Switch the buffers, put the used one at the end of container of buffers,
    /// the last one will be then reused    
    if (rep_num_buffer != last_ind)
    {
      void *buf_tmp = buffers[rep_num_buffer];
      buffers[rep_num_buffer] = buffers[last_ind];
      buffers[last_ind] = buf_tmp;
      working_buf = buf_tmp;
    }

    /// Reset the initial buffer parameters and increment number of buffers
    cur_buf = (void*)((char*)working_buf + 2*INT_SIZE);
    buf_size = 2*INT_SIZE;
    num_blocks = 0;
    num_buffers++;
  }
}

/// Attach message to the buffer
int MSG_Buffer_Handle::attach_data(void *attach_buf, int size_attach_buf, int fixed_size)
{ 
  /// If no buffers allocated yet, allocate the first one
  if(!num_buffers)
  {
    alloc();
  }

  /// Estimated size of buffer after message is added
  int new_buf_size = buf_size + size_attach_buf;

  /// If the message size is not fixed, add the size of integer which tells the size of the message
  if (!fixed_size)
  {
    new_buf_size += INT_SIZE;
  }

  /// If buffer is full
  if (new_buf_size > max_size)
  {
    /// Finilize the buffer
    wrap_buffer(0);
    return BUF_FULL;
  }
  else 
  {
    /// Put the message to the buffer
    if (!fixed_size)
    {
      memcpy(cur_buf, (void*)(&size_attach_buf), INT_SIZE);
      cur_buf = (void*)((char*)cur_buf + INT_SIZE);
    }
    memcpy(cur_buf, attach_buf, size_attach_buf);
    cur_buf = (void*)((char*)cur_buf + size_attach_buf);
    buf_size = new_buf_size;
    num_blocks++;
    return buf_size;
  }
}

/// Extract the message from the buffer
int MSG_Buffer_Handle::get_data(void *&ret_buf, int fixed_size)
{
  int size_ret_buf, nb;
 
  /// If it's the beginning of the buffer
  if (cur_num_blocks == 0)
  {
    /// Get the number of messages
    num_blocks = *(int*)working_buf;
    cur_buf = (void*)((char*)working_buf + INT_SIZE);

    if (num_blocks == 0)
    {
      return BUF_EMPTY;
    }

    /// Try to get the total number of buffers, if it is 0, this is not the last buffer
    nb = *(int*)cur_buf;
    cur_buf = (void*)((char*)cur_buf + INT_SIZE);
    if (nb)
    {
      final_num_buffers = nb;
    }
  }

  /// If not all the messages were extracted from the buffer
  if (cur_num_blocks < num_blocks)
  {
    /// If the message size is arbitrary, read the size of the message, and then proceed to the message itself
    if (!fixed_size)
    {
      size_ret_buf = *(int*)cur_buf;
      cur_buf = (void*)((char*)cur_buf + INT_SIZE);
    }
    else
    {
      size_ret_buf = fixed_size;
    }
  
    /// Extract the message
    ret_buf = malloc(size_ret_buf);
    memcpy(ret_buf, cur_buf, size_ret_buf);
    cur_buf = (void*)((char*)cur_buf + size_ret_buf);    
    cur_num_blocks++;
    return size_ret_buf;
  }
  else
  {
    return BUF_READ;
  }
}

/// Finalize the last buffer
void MSG_Buffer_Handle::finalize_buffer()
{
  /// If there was at least one buffer, put the total number inside
  if (num_buffers)
  {
    wrap_buffer(num_buffers);
  }
  else
  {
    /// Create an empty buffer
    working_buf = malloc(INT_SIZE);
    *(int*)working_buf = NO_BUFFERS;
    buf_size = INT_SIZE;
  }
}

/// Find available buffer which can be used for receiving package of messages
void *MSG_Buffer_Handle::get_buffer_to_recv()
{
  void *buf;

  /// If there is at least buffer which can be reused, take it
  if (read_num_buf_diff > 0)
  {
    buf = buffers[replace_num_buffer];
    replace_num_buffer = (replace_num_buffer + 1) % num_active_buffers;
    read_num_buf_diff--;
  }
  /// If there is only one buffer or no buffers at all
  else if (cur_num_buffer == 0)
  {
      buf = malloc(max_size);
      buffers.push_back(buf);
      num_active_buffers++;
  }
  /// Allocate the buffer and put it before cur_num_buffer to keep the ring of buffers in order
  else
  {
      buf = malloc(max_size);
      buffers.insert(buffers.begin() + cur_num_buffer, buf);
      cur_num_buffer++;     
      replace_num_buffer++;
      num_active_buffers++;
  }
  /// If it is the first buffer, assign working buffer to NULL
  if (num_buffers == 0)
  {
    working_buf = NULL;
  }
  num_buffers++;
  return buf;
}

/// Try to get the next available buffer to receive a package
int MSG_Buffer_Handle::get_next_rcvd_buffer(int rep_num_buffer)
{
  if (num_buffers)
  {
    int try_num_buf_diff = read_num_buf_diff + 1;

    /// All the existent buffers might be already read, then no buffer for extracting messages is available
    if (try_num_buf_diff < num_active_buffers)
    {
      cur_num_buffer = (cur_num_buffer + 1) % num_active_buffers;
      read_num_buf_diff = try_num_buf_diff;

      assign_working_buffer(rep_num_buffer);

      return 1;
    }
  }
  return 0;
}

/// Assign working buffer based on the number of available buffer
void MSG_Buffer_Handle::assign_working_buffer(int rep_num_buffer)
{
  /// If no buffer is available
  if (rep_num_buffer == -1)
  {
    working_buf = NULL;
  }
  else
  {
    /// The index of the buffer which is available
    int ind_buf = (cur_num_buffer + rep_num_buffer) % num_active_buffers;

    /// Replace the available buffer with the one which is at the beginning of the ring of buffers (cur_num_buffer)
    if (ind_buf != cur_num_buffer)
    {
      void *buf_tmp = buffers[ind_buf];
      buffers[ind_buf] = buffers[cur_num_buffer];
      buffers[cur_num_buffer] = buf_tmp;
    }
    working_buf = buffers[cur_num_buffer];
    cur_num_blocks = 0;
  }
}

/*
void MSG_Buffer_Handle::take_next_buffer()
{
  if (num_buffers)
  {
    cur_num_buffer = (cur_num_buffer + 1) % num_active_buffers;
  }
  working_buf = buffers[cur_num_buffer];
  cur_buf = (void*)((char*)working_buf + 2*INT_SIZE);
  buf_size = 2*INT_SIZE;
  num_blocks = 0;
  num_buffers++;
}

int MSG_Buffer_Handle::take_next_sent_buffer()
{
  if (cur_num_buffer < num_active_buffers)
  {
    cur_num_buffer++;
    working_buf = buffers[cur_num_buffer];
    cur_buf = (void*)((char*)working_buf + 2*INT_SIZE);
    buf_size = 2*INT_SIZE;
    num_blocks = 0;
    num_buffers++;
    return 1;
  }
  else
  {
    return 0;
  }
}
*/
