/******************************************************************************

  (c) 2009-2011 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/

#include "IPComMan.h"

/// Static Initialization of IPComMan
IPComMan::IPComMan()
{
  myrank = 0;
  mysize = 0;

  tag = 0;
  comm_val = Neighbors;

  /// Set the pid the receiving part begins from
  last_used_pid = -1;

  num_probes = 0;
  num_received = 0;

  fixed_msg_size = 0;
  buf_sent = buf_recvd = 0;
  max_buf_size = 10240;
}

IPComMan::IPComMan(MPI_Comm comm, int set_tag, int max_buffer_size, Comm_Validation cval):
  max_buf_size(max_buffer_size + 2*INT_SIZE), tag(set_tag), comm_val(cval)
{
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &mysize);

  int res = MPI_Comm_dup(comm, &local_comm);
  MPI_Errhandler_set(local_comm, MPI_ERRORS_RETURN);

  /// Set the pid the receiving part begins from
  last_used_pid = -1;

  num_probes = 0;
  num_received = 0;

  fixed_msg_size = 0;
  buf_sent = buf_recvd = 0;
}

void IPComMan::init(MPI_Comm comm, int set_tag, int max_buffer_size, Comm_Validation cval)
{
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &mysize);

  int res = MPI_Comm_dup(comm, &local_comm);
//  local_comm = comm;
  MPI_Errhandler_set(local_comm, MPI_ERRORS_RETURN);

  max_buf_size = max_buffer_size + 2*INT_SIZE;
  tag = set_tag;
  comm_val = cval;

  /// Set the pid the receiving part begins from
  last_used_pid = -1;

  num_probes = 0;
  num_received = 0;

  fixed_msg_size = 0;
  buf_sent = buf_recvd = 0;
}

/// Initialization of IPComMan with neighborhood
IPComMan::IPComMan(MPI_Comm comm, std::set<int>& nbr_parts, int set_tag, int max_buffer_size, Comm_Validation cval):
  max_buf_size(max_buffer_size + 2*INT_SIZE), tag(set_tag), comm_val(cval)
{
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &mysize);

  int res = MPI_Comm_dup(comm, &local_comm);
//  local_comm = comm;
  MPI_Errhandler_set(local_comm, MPI_ERRORS_RETURN);

  /// Set the pid the receiving part begins from
  last_used_pid = -1;

  num_probes = 0;
  num_received = 0;

  fixed_msg_size = 0;
  buf_sent = buf_recvd = 0;
  set_Nbrs(nbr_parts);
}

/// Setting the neighbors
void IPComMan::set_Nbrs(std::set<int>& nbr_parts)
{
  Nbrs.clear();

  /// Excluding mypid from the set of neighbors, if it is there
  if (nbr_parts.size() > 0)
  {
    std::set<int>::iterator it_nbr_parts = nbr_parts.find(myrank);
    if (it_nbr_parts != nbr_parts.end())
    {
      nbr_parts.erase(it_nbr_parts);
    }

    it_nbr_parts = nbr_parts.begin();

    for (; it_nbr_parts != nbr_parts.end(); it_nbr_parts++)
    {
      if (*it_nbr_parts != myrank)
      {
        Nbrs.insert(*it_nbr_parts);
      }
    }
  }
}

/// Setting the neighbors equal to the size of the whole domain
void IPComMan::set_Nbrs_All()
{
  Nbrs.clear();
  for (int i = 0; i < mysize; i++)
  {
    Nbrs.insert(i);
  }
  Nbrs.erase(myrank);
}

/*
void IPComMan::init(MPI_Comm comm)
{
  MPI_Comm_rank(comm, &myrank);
  MPI_Comm_size(comm, &mysize);

  int res = MPI_Comm_dup(comm, &local_comm);
  MPI_Errhandler_set(local_comm, MPI_ERRORS_RETURN);
}

int IPComMan::init_comm_plan(set<int>* nbr_parts, int set_tag)
{
  set<int>::iterator it_nbr_parts = nbr_parts->find(myrank);
  if (it_nbr_parts != nbr_parts->end())
  {
    nbr_parts->erase(it_nbr_parts);
  }

  it_nbr_parts = nbr_parts->begin();
  last_used_pid = *it_nbr_parts;

  for (it_nbr_parts; it_nbr_parts != nbr_parts->end(); it_nbr_parts++)
  {
    if (*it_nbr_parts != myrank)
    {
      Nbrs.insert(*it_nbr_parts);
    }
  }

  num_received = 0;
//  tag = (tag+1)%1024; 
  tag = set_tag;

  return tag;
}
*/

/// Send message with fixed message size
void IPComMan::send(int pid, void* msg)
{
  send(pid, msg, fixed_msg_size);
}

/// Send message with arbitrary size
void IPComMan::send(int pid, void* msg, int msg_size)
{
  /// Message buffer(s) corresponding to each pid
  MSG_Buffer_Handle *bh;

  /// Container of MPI_Requests corresponding to each buffer for a specific pid
  req_container *reqs;

  msg_handler msh;
  int buf_status;

  /// Identify, whether there's an entry for this pid in the map of pids and buffers
  mob_iter it_mosends = mosends.find(pid);
  if (it_mosends == mosends.end())
  {
    /// If it is a new neighbor
    if (!Nbrs.count(pid))
    {
//      Nbrs.insert(pid);
      /// In sending routine put a new neighbor in the separate container, otherwise the program might stuck during receiving phase
      /// waiting for the message from this new neighbor, which maybe unaware of mypid yet
      new_Nbrs.insert(pid);

      MPI_Request *sreq = new MPI_Request;
      int *send_int = (int*)malloc(sizeof(int));
      *send_int = NEW_NEIGHBOR;

      //printf("[%d: %d to %d]: Sending info to a new pid in the neighborhood\n", tag, myrank, pid);
   
      /// Send information about joining mypid to the neighborhood of pid immediately
      MPI_Isend(send_int, 1, MPI_INTEGER, pid, tag, local_comm, sreq);

      /// Save a request for the message sent
      sys_reqs[(void*)send_int] = sreq;

      msg_sent[pid]++;
    }

    /// create an object of message buffer(s) and MPI requests for those buffers
    bh = new MSG_Buffer_Handle(max_buf_size);
    reqs = new req_container;
    msh.msg_buffers = bh;
    msh.req_handles = reqs;
    mosends[pid] = msh;
  }
  else
  {
    /// take existent ones
    msh = it_mosends->second;
    bh = msh.msg_buffers;
    reqs = msh.req_handles;
  }

  /// Identify whether the buffer is full after attaching the message
  buf_status = bh->attach_data(msg, msg_size, fixed_msg_size);

  /// If it's full, send it and create another one
  if (buf_status == BUF_FULL)
  {
    MPI_Request *request = new MPI_Request;
    MPI_Isend(bh->get_buffer(), bh->get_buf_size(), MPI_BYTE, pid, tag, local_comm, request);

    /// Figure out if one of the requests is satisfied
    int num_req = get_free_req_number_send(reqs, bh->get_num_active_buffers(), request);

    /// Allocate new buffer or use the existent one if it was completely sent, based on the information from requests
    bh->get_buffer_to_send(num_req);    

    buf_sent++;

    bh->attach_data(msg, msg_size, fixed_msg_size);
  }
  msg_sent[pid]++; 
}

/// Close all the buffers and send all the packages
void IPComMan::finalize_send()
{
  MSG_Buffer_Handle *bh;
  req_container *reqs;
  msg_handler msh;
  mob_iter it_mosends;
  int pid;

  /// Loop through the neighbors, close all the buffers, put the total number of buffers into last buffer, and send
  for (std::set<int>::iterator it_nbrs = Nbrs.begin(); it_nbrs != Nbrs.end(); it_nbrs++)
  {
    pid = *it_nbrs;
    it_mosends = mosends.find(pid);
    
    /// If there was no messages for pid in the neighborhood, send the constant stating that the communication is over
    if (it_mosends == mosends.end())
    {
      bh = new MSG_Buffer_Handle(max_buf_size);
      reqs = new req_container;
      msh.msg_buffers = bh;
      msh.req_handles = reqs;
      mosends[pid] = msh;
      msg_sent[pid]++;
    }
    else
    {
      msh = it_mosends->second;
      bh = msh.msg_buffers;
      reqs = msh.req_handles;
    }
  
    /// Close the buffer
    bh->finalize_buffer();

    MPI_Request *request = new MPI_Request;
    MPI_Isend(bh->get_buffer(), bh->get_buf_size(), MPI_BYTE, pid, tag, local_comm, request);

    /// Add request to the list of requests
    add_request(reqs, bh->get_num_active_buffers(), request);
//    get_free_req_number_send(reqs, bh->get_num_active_buffers(), request);

    buf_sent++;
  }

  /// Same for the new neighbors
  for (std::set<int>::iterator it_nbrs = new_Nbrs.begin(); it_nbrs != new_Nbrs.end(); it_nbrs++)
  {
    pid = *it_nbrs;
    it_mosends = mosends.find(pid);
    if (it_mosends == mosends.end())
    {
      bh = new MSG_Buffer_Handle(max_buf_size);
      reqs = new req_container;
      msh.msg_buffers = bh;
      msh.req_handles = reqs;
      mosends[pid] = msh;
      msg_sent[pid]++;
    }
    else
    {
      msh = it_mosends->second;
      bh = msh.msg_buffers;
      reqs = msh.req_handles;
    }
    bh->finalize_buffer();
    MPI_Request *request = new MPI_Request;
    MPI_Isend(bh->get_buffer(), bh->get_buf_size(), MPI_BYTE, pid, tag, local_comm, request);

    add_request(reqs, bh->get_num_active_buffers(), request);
//    get_free_req_number_send(reqs, bh->get_num_active_buffers(), request);

    buf_sent++;
  }

/*
  /// hack
  for (map<int,int>::iterator mit = msg_sent.begin(); mit != msg_sent.end(); mit++)
  {
    printf("[%d: %d to %d]: sent %d elements\n", tag, myrank, mit->first, mit->second);
  }
  /// hack;
*/
}

/// Receiving phase
int IPComMan::receive(void *&msg, int *pid_from)
{
  int possibly_done_comm = 1;

  /// There were situations when after receiving the last message this condition should be checked in the begining, now not sure 
  /// whether it's needed or not, probably useful for having no neighbors at all, but can be avoided and checked later anyway
  if (num_received == Nbrs.size())
  {
    if (comm_val == All_Reduce)
    {
      if (comm_done())
      {
        finalize_receive();
        return ALL_RECVD;
      }
      else
      {
        possibly_done_comm = 0;
        last_used_pid = -1;
      }
    }
    else
    {
      finalize_receive();
      return ALL_RECVD;
    }
  }

  /// Initial size of the receiving message
  int rc_size = -1;

  MSG_Buffer_Handle *bh;
  req_container *reqs;
  msg_handler msh;
  MPI_Status status;
  int pid, flag, count;  
  mob_iter it_morecvs;

  while(1)
  {
    /// Check for the arrival of the buffer
    MPI_Iprobe(MPI_ANY_SOURCE, tag, local_comm, &flag, &status);

    /// If the container of pids and their received buffers is empty, or all the buffers were read
    if (morecvs.size() == 0 || rc_size == 0 && flag == 0)
    {
      /// If MPI_Allreduce with global sends and receives was not satisfied, 
      /// i.e. if the number of global sends and receives is not equal
      if(possibly_done_comm == 0)
      {
        possibly_done_comm = 1;
      }
      else
      {
        MPI_Probe(MPI_ANY_SOURCE, tag, local_comm, &status);
        flag = 1;
      }
    } 

    /// While there are arrived buffers in the queue to receive
    while (flag)
    {
      void *recv_buf;
      MPI_Get_count(&status, MPI_BYTE, &count);
      buf_recvd++;

      MPI_Status l_status;

      pid = status.MPI_SOURCE;
 
      /// If buffer size is equal to 4
      if (count == INT_SIZE)
      {
        int recv_int;
        MPI_Recv(&recv_int, 1, MPI_INTEGER, pid, tag, local_comm, &l_status);

        /// If it is a new neighbor, add it to the set of neighbors
        if (!Nbrs.count(pid))
        {
          Nbrs.insert(pid);
        }

        /// If it is a constant specifying that there will be no buffers from the current pid
        if (recv_int == NO_BUFFERS)
        {
          num_received++;
        }
        /// Otherwise if it is a constant telling that this pid is going to be in the neighborhood
        else if (recv_int == NEW_NEIGHBOR)
        {
          /// If it is a new neighbor, to whom nothing was sent, send him a message that there will be no communication,
          /// otherwise pid who sent something to myrank will not add myrank to list of his neighbors
          if (!new_Nbrs.count(pid)) 
          {
            MPI_Request *sreq = new MPI_Request;
            int *send_int = (int*)malloc(sizeof(int));
            *send_int = -1;
            MPI_Isend(send_int, 1, MPI_INTEGER, pid, tag, local_comm, sreq);
            sys_reqs[(void*)send_int] = sreq;
            msg_sent[pid]++;
          }
        }
        msg_recvd[pid]++;
      }
      else
      {
        // Adding buffer to the map of pids <-> buffers
        it_morecvs = morecvs.find(pid);
        if (it_morecvs == morecvs.end())
        {
          /// If the buffer from pid who is not in the neighborhood arrived before the notification of new neighbor
          if (!Nbrs.count(pid))
          {
            Nbrs.insert(pid);
          }
          bh = new MSG_Buffer_Handle(max_buf_size);
          reqs = new req_container;
          msh.msg_buffers = bh;
          msh.req_handles = reqs;
          morecvs[pid] = msh;
        }
        else
        {
          msh = it_morecvs->second;
          bh = msh.msg_buffers;
          reqs = msh.req_handles;
        }
        recv_buf = bh->get_buffer_to_recv();
        if (last_used_pid < 0 || last_used_pid > pid)
        {
          last_used_pid = pid;
        }

        MPI_Request *request = new MPI_Request;
        MPI_Irecv(recv_buf, count, MPI_BYTE, pid, tag, local_comm, request);
        reqs->push_back(request);
        
      }
      /// Check to see if there is more buffers ready to be received
      MPI_Iprobe(MPI_ANY_SOURCE, tag, local_comm, &flag, &status);
    }

    /// Find pid from which buffer the last message was extracted
    it_morecvs = morecvs.find(last_used_pid);  
    while (it_morecvs != morecvs.end())
    {
      pid = it_morecvs->first;
      msh = it_morecvs->second;
      bh = msh.msg_buffers;
      reqs = msh.req_handles;      

      /// Check if it's the first buffer to be proceeded
      if (bh->get_buffer() == NULL && reqs->size())
      {
        /// Try to get the number of buffer which was completely received
        int num_req = get_free_req_number_recv(reqs);

        /// Pik up the buffer based on its number, if available
        bh->assign_working_buffer(num_req);
        if (num_req > -1)
        {
          delete (*reqs->begin());
          reqs->erase(reqs->begin());
        }
        else
        {
          /// There is no available buffer for the current pid, proceed to the next pid
          it_morecvs++;
          rc_size = -2;
          continue;
        }
      }

      /// Extract message from the buffer
      rc_size = bh->get_data(msg, fixed_msg_size);

      /// If the buffer is read to the end
      if (rc_size == BUF_READ)
      {
        /// Check whether all the buffers have been received and processed
        if (bh->get_num_buffers() == bh->get_final_num_buffers() && reqs->size() == 0)
        {
          mob_iter it_del = it_morecvs;

          it_morecvs++;
          delete bh;
          delete reqs;
          morecvs.erase(it_del);
          num_received++;

          /// If it's the end of the map, continue with its begining
          if (it_morecvs == morecvs.end())
          {
            it_morecvs = morecvs.begin();
            continue;
          }          
          else
          {
            last_used_pid = it_morecvs->first;
          }
        }

        else
        {
          if (bh->get_num_active_buffers() > 1)
          {
            /// Get the available buffer
            int num_req = get_free_req_number_recv(reqs);

            /// Try to get the next buffer available for being read. There may be all buffers processed which were received
            /// until now, but there still more buffers coming from the current pid
            int next_buf_status = bh->get_next_rcvd_buffer(num_req);
            if (next_buf_status)
            {
              /// If buffer is available, extract message from it
              if (num_req > -1)
              {
                rc_size = bh->get_data(msg, fixed_msg_size);
                delete (*reqs->begin());
                reqs->erase(reqs->begin());
                last_used_pid = pid; 
                break;
              }
              else
              {  
                it_morecvs++;
                rc_size = -2;
                continue;
              }
            }
          }
          it_morecvs++;
        }
      }
      else
      {
        break;
      }
    }

    /// If the map of pid <-> received buffers is empty
    if (morecvs.size() == 0)
    {
      /// If the number of pids from which all the messages were received equal to the number of my neighbors
      if (num_received == Nbrs.size())
      {
        if (comm_val == All_Reduce)
        {
          /// Check the global amount of sent and received messages
          if (comm_done())
          {
            /// Clean up all the sends and receives
            finalize_receive();
            return ALL_RECVD;
          }
          else
          {
            possibly_done_comm = 0;
            last_used_pid = -1;
          }
        }
        else
        {
          finalize_receive();
          return ALL_RECVD;
        }
      }
      else
      {
        /// Continue to receive the buffers and extract messages since there are still some pids left
        rc_size = 0;
        last_used_pid = -1; 
        continue;
      }
    } 
    else
    {
      /// If no message was extracted, but there are still pids from which buffers should arrive, 
      /// or if there was no available buffer to extract message from
      if (rc_size == 0 || rc_size == -2)
      {
        last_used_pid = morecvs.begin()->first;
        continue;
      }
      else
      {
        /// Deliver the message to user
        *pid_from = pid; 
        last_used_pid = pid;
        
        msg_recvd[pid]++;   

        break;
      }
    }
  }
  return rc_size;
}

/// Clean up all the maps for sends and receives, and wait for all the messages to be sent
void IPComMan::finalize_receive()
{
  int flag;
  MPI_Status status;

  /// Wait for all maintenance sends to be completed
  while (sys_reqs.size())
  {
    for (sys_map_iter sys_it = sys_reqs.begin(); sys_it != sys_reqs.end(); sys_it++)
    {
      MPI_Test(sys_it->second, &flag, &status);
      if (flag)
      {
        free(sys_it->first);
        delete(sys_it->second);
        sys_reqs.erase(sys_it);
      }
    }
  }

  /// Wait for all buffer sends to be completed
  MSG_Buffer_Handle *bh;  
  req_container *reqs;
  for (mob_iter it_mosends = mosends.begin(); it_mosends != mosends.end(); it_mosends++)
  {
    reqs = it_mosends->second.req_handles;
    while(reqs->size())
    {
      it_req_container it_reqc = reqs->begin();
      int it = 0;
      while(1)
      {
        MPI_Request *request = reqs->at(it);
        MPI_Test(request, &flag, &status);
        if (flag)
        {
          delete(request);
          reqs->erase(it_reqc + it);
        }
        else
        {
          it++;
        }
        if (it == reqs->size())
          break;
      }
    }
    delete reqs;
    bh = it_mosends->second.msg_buffers;
    delete bh;
  }
  mosends.clear();
  morecvs.clear();

  /// Clean the list of new neighbors
  new_Nbrs.clear();

  /// Reset the last pid from which the buffer was read
  last_used_pid = -1;
  num_probes = 0;

  /// Set the number of processed pids to 0
  num_received = 0;

  buf_sent = 0;
  buf_recvd = 0;

#ifdef DEBUG
  int num_sent = 0, num_recvd = 0;
  for (std::map<int,int>::iterator mit = msg_sent.begin(); mit != msg_sent.end(); ++mit)
  {
    num_sent += mit->second;
  }
  for (std::map<int,int>::iterator mit = msg_recvd.begin(); mit != msg_recvd.end(); ++mit)
  {
    num_recvd += mit->second;
  }
  printf("Pid [%d] sent %d buffers, received %d buffers\n", myrank, num_sent, num_recvd);
#endif

  /// Clear number of sent and received messages
  msg_sent.clear();
  msg_recvd.clear();
}

/// Add current request to the list of requests for current pid
void IPComMan::add_request(req_container *reqs, int num_active_buffers, MPI_Request *request)
{
  /// The current request is always put at the end of the container
  if (reqs->size() < num_active_buffers || num_active_buffers == 0)
  {
    reqs->push_back(request);
  }
  else
  {
    /// Free the request which was satisfied and put the current one on its place, the last element in the request container
    delete(reqs->at(reqs->size()-1));
    reqs->at(reqs->size()-1) = request;
  }
}

/// Get the number of the buffer which can be reused for the next set of messages based on the request information
int IPComMan::get_free_req_number_send(req_container *reqs, int num_active_buffers, MPI_Request *request)
{
  add_request(reqs, num_active_buffers, request);

  int flag;
  MPI_Status status;
  int num_req = -1;
  int num_req_free = -1;
  for (it_req_container it_reqs = reqs->begin(); it_reqs != reqs->end(); it_reqs++)
  {
    num_req_free++;
    MPI_Test(*it_reqs, &flag, &status);
    if (flag)
    {
      num_req = num_req_free;
      int last_ind = reqs->size() - 1;
   
      /// Switch the request which is satisfied with the last one in the container, 
      /// the last one will be then freed and substituted with the new one
      if (num_req != last_ind)
      {
        MPI_Request *req_tmp = reqs->at(num_req);
        reqs->at(num_req) = reqs->at(last_ind);
        reqs->at(last_ind) = req_tmp;
      }
      break;
    }
  }
  return num_req;
}

/// Get the number of the buffer from which the message can be extracted
int IPComMan::get_free_req_number_recv(req_container *reqs)
{
  int flag;
  MPI_Status status;
  int num_req = -1;
  int num_req_free = -1;
  for (it_req_container it_reqs = reqs->begin(); it_reqs != reqs->end(); it_reqs++)
  {
    num_req_free++;
    MPI_Test(*it_reqs, &flag, &status);
    if (flag)
    {
      num_req = num_req_free;

      /// Switch the request which is satisfied with the first one in the container,
      /// the first one will be then erased
      if (num_req != 0)
      {
        MPI_Request *req_tmp = reqs->at(num_req);
        reqs->at(num_req) = reqs->at(0);
        reqs->at(0) = req_tmp;
      }
      break;
    }
  }
  return num_req;
}

/// Check for global number of sends and receives
int IPComMan::comm_done()
{
  int local_sent_recvd[2] = {0, 0};
  int global_sent_recvd[2];
  for (std::map<int,int>::iterator mit = msg_sent.begin(); mit != msg_sent.end(); mit++)
  {
    local_sent_recvd[0] += mit->second;
  }  
  for (std::map<int,int>::iterator mit = msg_recvd.begin(); mit != msg_recvd.end(); mit++)
  {
    local_sent_recvd[1] += mit->second;
  }

#ifdef DEBUG
  printf("[%d: %d]: Finishing communication... sent %d, recvd %d\n", tag, myrank, local_sent_recvd[0], local_sent_recvd[1]);
#endif

  MPI_Allreduce(local_sent_recvd, global_sent_recvd, 2, MPI_INTEGER, MPI_SUM, local_comm);
  if (global_sent_recvd[0] == global_sent_recvd[1])
  {
#ifdef DEBUG
    if (myrank == 0)
      printf("!!!!!Global sends and Global recvs are equal!!!!!!\n");
#endif
    return 1;
  }
  else
  {
#ifdef DEBUG
    if (myrank == 0)
    {
      printf("!!!!!Sends and recvs are not equal!!!!!! %d sends, %d recvs\n",global_sent_recvd[0], global_sent_recvd[1]);
    }
#endif
    return 0;
  }
}


/// Send the same message to every neighbor
int IPComMan::send_to_nbrs(void *msg, int msg_size)
{
  int pid;
  for (std::set<int>::iterator it_nbrs = Nbrs.begin(); it_nbrs != Nbrs.end(); it_nbrs++)
  {
    pid = *it_nbrs;
    send(pid, msg, msg_size);
  }
  return Nbrs.size();
}


/// Send the same message with the fixed size to every neighbor
int IPComMan::send_to_nbrs(void *msg)
{
  int pid;
  for (std::set<int>::iterator it_nbrs = Nbrs.begin(); it_nbrs != Nbrs.end(); it_nbrs++)
  {
    pid = *it_nbrs;
    send(pid, msg);
  }
  return Nbrs.size();
}


/// Set up neighbors of neighbors
int IPComMan::set_Nbrs_of_Nbrs()
{
  Nbrs_of_Nbrs.clear();

  set_comm_validation(Neighbors);
  set_tag(0);
  set_fixed_msg_size(sizeof(int));
  int* msg_send = (int*)alloc_msg(sizeof(int));
  int num_sent = 0, num_recvd = 0;
  
  for (std::set<int>::iterator it_nbrs = Nbrs.begin(); it_nbrs != Nbrs.end(); it_nbrs++)
  {
    for (std::set<int>::iterator it_nbrs_out = Nbrs.begin(); it_nbrs_out != Nbrs.end(); it_nbrs_out++)
    {
      if (*it_nbrs != *it_nbrs_out)
      {
        *msg_send = *it_nbrs_out;
        send(*it_nbrs, (void*)msg_send);
        num_sent++;
      }
    }
  }
  finalize_send();
  free_msg(msg_send);

  void *msg_recv;
  int pid_from;
  while (int rc = receive(msg_recv, &pid_from))
  {
    num_recvd++;
    Nbrs_of_Nbrs[pid_from].insert(*(int*)msg_recv);
    free_msg(msg_recv);
  }

  return Nbrs_of_Nbrs.size();
}
