/******************************************************************************

  (c) 2009-2011 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/

#ifndef _H_IPCOMMAN
#define _H_IPCOMMAN

#include "MSG_Buffer_Handle.h"

#include "mpi.h"
#include <memory>
#include <set>
#include <map>
#include <stdlib.h>

#define ALL_RECVD 0
#define MAX_PID 1000000

class IPComMan
{
  private:
    enum {NEW_NEIGHBOR = -2, MAX_NUM_PROBES = 1000};

    int max_buf_size, fixed_msg_size;
    int myrank, mysize, tag;

    /// The number of processed pids after receiving all the messages
    int num_received;

    /// The pid from which buffer in the receiving part the message is trying to be extracted 
    int last_used_pid;

    int num_probes;

    /// Local communicator
    MPI_Comm local_comm;  

    /// Set of neighbors  
    std::set<int> Nbrs;

    /// Set of pids becoming new neighbors during communication step
    std::set<int> new_Nbrs;

    /// Map for neighbors of neighbors
    std::map<int, std::set<int> > Nbrs_of_Nbrs;

    /// Number of send and received buffers
    int buf_sent, buf_recvd; 

    /// Maps of pid <-> number of messages, for sends and receives
    std::map<int,int> msg_sent, msg_recvd;

    /// Request container
    typedef std::vector<MPI_Request *> req_container;
    typedef std::vector<MPI_Request *>::iterator it_req_container;

    /// Structure for buffers and correspondent requests
    typedef struct
    {
      MSG_Buffer_Handle *msg_buffers;
      req_container *req_handles;
    } msg_handler;
 
    /// Map of pid <-> buffers (with container of requests)
    typedef std::map<int, msg_handler> map_of_buffers;
    typedef map_of_buffers::iterator mob_iter;
//    typedef map<int, map_of_buffers> s_map_of_plans;
    typedef std::map<void*, MPI_Request*> sys_map_of_reqs;
    typedef std::map<void*, MPI_Request*>::iterator sys_map_iter;
   
    /// Maps of pid <-> buffers, for sending buffers and receiving ones 
    map_of_buffers mosends, morecvs;
 
    /// Map of message <-> request, for maintenance messages
    sys_map_of_reqs sys_reqs;


  protected:
    /// Clean up all the maps for sends and receives, and wait for all the messages to be sent
    void finalize_receive();

    /// Add current request to the list of requests for current pid
    void add_request(req_container *reqs, int num_active_buffers, MPI_Request *request);

    /// Get the number of the buffer which can be reused for the next set of messages based on the request information
    int get_free_req_number_send(req_container *reqs, int num_active_buffers, MPI_Request *request);

    /// Get the number of the buffer from which the message can be extracted
    int get_free_req_number_recv(req_container *reqs);

    /// Check for global number of sends and receives
    int comm_done();


  public:

    /// Method to check-up the end of communication round - either based on number of neighbors, or global number of messages
    /// The first one is valid when the neighborhood does not change throuout the communication round
    enum Comm_Validation {Neighbors, All_Reduce} comm_val;

    /// Setting the neighbors
    void set_Nbrs(std::set<int>& nbr_parts);

    /// Setting the neighbors equal to the size of the whole domain
    void set_Nbrs_All();

    /// Send message with fixed message size 
    void send(int pid, void* msg);

    /// Send the same message to every neighbor
    int send_to_nbrs(void *msg, int msg_size);

    /// Send the same message with the fixed size to every neighbor
    int send_to_nbrs(void *msg);

    /// Send message with arbitrary size
    void send(int pid, void* msg, int msg_size);

    /// Close all the buffers and send all the packages
    void finalize_send();

    /// Receive the message
    int receive(void *&msg, int *pid_from);

    /// Set up neighbors of neighbors
    int set_Nbrs_of_Nbrs();
    
    /// Get the current tag
    inline int get_tag(void){ return tag; }

    /// Set the tag
    inline void set_tag(int stag){ if(!stag) tag = (tag+1)%32768; else tag = stag; }
//    inline void set_tag(int stag){ if(!stag) tag = (tag+1)%32768; else tag = stag; printf("[%d]: PID %d has %d nbrs\n",tag,myrank,Nbrs.size()); }

    /// Number of neighbors
    inline int get_Num_Nbrs() { return Nbrs.size(); }

    /// Set the constant message size
    inline void set_fixed_msg_size(int msg_size){ fixed_msg_size = msg_size; }

    /// Get the constant message size
    inline int get_fixed_msg_size(void){ return fixed_msg_size; }

    /// Set the method of the end of communication round check-up
    inline void set_comm_validation(Comm_Validation cval){ comm_val = cval; }

    /// Get the method of the end of communication round check-up
    inline Comm_Validation get_comm_validation(void){ return comm_val; }

    /// Set the maximum buffer size
    inline void set_buf_size(int buf_size){ max_buf_size = buf_size; }

    /// Get the maximum buffer size
    inline int get_buf_size(void){ return max_buf_size; }

    /// Allocate msg_size bytes for the message
    inline void *alloc_msg(int msg_size){ return malloc(msg_size); }

    /// Free the message
    inline void free_msg(void *msg){ free(msg); }

    /// Check if the requested pid is a neighbor
    inline int is_Nbr(int pid){ return Nbrs.count(pid); }

    /// Check if the requested nbrof is a neighbor of pid
    inline int is_Nbr_of_Nbr(int pid, int nbrof){ return Nbrs_of_Nbrs[pid].count(nbrof); }    

    /// Initialization of IPComMan
    IPComMan();
    IPComMan(MPI_Comm comm, int set_tag, int max_buffer_size, Comm_Validation cval);
    void init(MPI_Comm comm, int set_tag, int max_buffer_size, Comm_Validation cval);

    /// Initialization of IPComMan with neighborhood
    IPComMan(MPI_Comm comm, std::set<int>& nbr_parts, int set_tag, int max_buffer_size, Comm_Validation cval);
};

#endif
